
// (c) COPYRIGHT URI/MIT 1996
// Please read the full copyright statement in the file COPYRIGH.  
//
// Authors:
//	reza		Reza Nekovei (reza@intcomm.net)

// DSP sub-class implementation for DSPByte,..DSPGrid.
// The files are patterned after the subcalssing examples 
// Test<type>.c,h files.
//
// ReZa 8/25/96

// $Log: DSPArray.cc,v $
// Revision 1.4  1998/08/06 16:32:06  jimg
// Fixed misuse of the read(...) member function. Return true if more data
// is to be read, false is if not and error if an error is detected
//
// Revision 1.3  1996/11/15 18:22:12  jimg
// Fixed copyright statement.
// Added include of config header
//
// Revision 1.2  1996/10/15 16:14:28  reza
// Fixed the stride counter to get subsamples (not subsections).
//
// Revision 1.1  1996/09/18 22:41:54  reza
//  First release of DODS-DSP servers.
//
//

#include "config_dsp.h"

static char rcsid[]={"$Id: DSPArray.cc,v 1.4 1998/08/06 16:32:06 jimg Exp $"};

#ifdef __GNUG__
#pragma implementation
#endif

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <iostream.h>
#include <assert.h>

#include "cgi_util.h"
#include "DSPArray.h"
#include "Dtype.h"

static char Msgt[255]; // Used by ErrMsgT

extern "C" int    OpenDsp(const char *filename);
extern "C" void   CloseDsp();
extern "C" void   cvtlng_(long *ilon, long *ilat, float *lon, float *lat);
extern "C" int    CheckImage(int *nline, int *npix);
extern "C" int    GetImageLine(char **buf, int LineNo);
extern "C" D_type VarType();

Array *
NewArray(const String &n, BaseType *v)
{
    return new DSPArray(n, v);
}

BaseType *
DSPArray::ptr_duplicate()
{
    return new DSPArray(*this);
}

DSPArray::DSPArray(const String &n, BaseType *v) : Array(n, v)
{
}

DSPArray::~DSPArray()
{
}

bool
DSPArray::read(const String &dataset, int &error)
{
    int start, stride, stop;
    int start_p, stride_p, stop_p;
    float lon, lat;
    int nline, npix; 
    dods_int32   *BufInt32;
    dods_float64 *BufFlt64;
    dods_byte    *dataBuf;

    if (read_p())  // Nothing to do
	return false;

    if (!OpenDsp(dataset)) {
	sprintf(Msgt, "DSPArray: Could not open file %s",(const char *)dataset);
	ErrMsgT(Msgt);
	error = 1;
	return false;
    }

    Pix p = first_dim();
    start = dimension_start(p,true);
    stride = dimension_stride(p, true);
    stop = dimension_stop(p, true); 

    if (!CheckImage(&nline, &npix)) {
	sprintf(Msgt, "DSPArray: Couldn't initialize image-input module");
	ErrMsgT(Msgt);
	error = 1;
	return false;
    } else 

	// Get lat values from DSP
	if (name() == "lat") {
	    if(start+stop+stride == 0){ //default
		start = 0;
		stride = 1;
		stop = nline-1;
	    }
	    dods_float64 *flt64 = new dods_float64 [((stop - start)/stride) + 1];
	    int count = 0;
	    long ilon = 0;
	    for (long ilat = start; ilat <= stop; ilat+=stride, count++) {
		cvtlng_(&ilon, &ilat, &lon, &lat);
		*(flt64+count) = (dods_float64) lat;
	    }
	    set_read_p(true);  
	    val2buf((void *)flt64);      
	    delete [] flt64; //clean up
	} else 

	    // Get lon values from DSP
	    if(name() == "lon") {
		if(start+stop+stride == 0){ //default
		    start = 0;
		    stride = 1;
		    stop = npix-1;
		}
		dods_float64 *flt64 = new dods_float64 [((stop - start)/stride) + 1];	
		int count = 0;
		long ilat = 0;
		for (long ilon = start; ilon <= stop; ilon+=stride, count++) {
		    cvtlng_(&ilon, &ilat, &lon, &lat);
		    *(flt64+count) = (dods_float64) lon;
		}
		set_read_p(true);  
		val2buf((void *)flt64);	
		delete [] flt64; //clean up
	    }
	    else {

		// Loop over number of lines (slower dimension) for actual data Array
		char    *buf; 
		next_dim(p);
		start_p = dimension_start(p,true);
		stride_p = dimension_stride(p, true);
		stop_p = dimension_stop(p, true); 

		if(start+stop+stride == 0){ //default lines
		    start = 0;
		    stride = 1;
		    stop = nline-1;
		}
		if(start_p+stop_p+stride_p == 0){ //default pixels
		    start_p = 0;
		    stride_p = 1;
		    stop_p = npix-1;
		}

		int Len = (((stop-start)/stride)+1)*(((stop_p-start_p)/stride_p)+1);

		D_type DType = VarType();
		D_type ConvertFrom =  D_UNSPECIFIED;

		switch (DType) { // allocate memory and set type conversion
		  case D_SHORT:
		    if(sizeof(short) != sizeof(dods_int32))
			ConvertFrom = D_SHORT;
		    BufInt32 = new dods_int32 [Len]; 
		    break;

		  case D_LONG:
		    if(sizeof(Dlong) != sizeof(dods_int32))
			ConvertFrom = D_LONG;
		    BufInt32 = new dods_int32 [Len]; 
		    break;
	    
		  case D_FLOAT:
		    if(sizeof(float) != sizeof(dods_float64))
			ConvertFrom = D_FLOAT;
		    BufFlt64 = new dods_float64 [Len]; 	
		    break;

		  case D_DOUBLE:
		    if(sizeof(double) != sizeof(dods_float64))
			ConvertFrom = D_DOUBLE;
		    BufFlt64 = new dods_float64 [Len]; 	
		    break;
	  
		  default:	
		    dataBuf = new dods_byte [Len]; 
		    break;
		}

		int Tcount = 0;
	
		if(ConvertFrom == D_UNSPECIFIED) { //typically no data conversion is needed
		    for (int iline = start; iline <= stop; iline+=stride) {	  
			if (!GetImageLine(&buf,iline)) {
			    sprintf(Msgt, "DSPArray: Error reading line %d from image",iline);
			    ErrMsgT(Msgt);
			    error = 1;
			    return false;
			}
			else{
			    for(int ipix = start_p; ipix <= stop_p; ipix+=stride_p){
				*(dataBuf+Tcount) = *(buf + ipix); 
				Tcount++;
			    }
			}
		    }
		    set_read_p(true);  
		    val2buf((void *)dataBuf);
		    delete [] dataBuf;
		}
		else { // Data needs type conversion
	  
		    for (int iline = start; iline <= stop; iline+=stride) {	  
			if (!GetImageLine(&buf,iline)) {
			    sprintf(Msgt, "DSPArray: Error reading line %d from image",iline);
			    ErrMsgT(Msgt);
			    error = 1;
			    return false;
			}
			else{
			    short  *bufsht = (short *) buf;
			    Dlong  *buflng = (Dlong *) buf;
			    float  *bufflt = (float *) buf;
			    double *bufdbl = (double *) buf;

			    for(int ipix = start_p; ipix <= stop_p; ipix+=stride_p){
		
				switch (ConvertFrom) {
				  case D_SHORT:
				    *(BufInt32+Tcount) = (dods_int32) *(bufsht+ipix);  
				    break;

				  case D_LONG:
				    *(BufInt32+Tcount) = (dods_int32) *(buflng+ipix);  
				    break;
		
				  case D_FLOAT:
				    *(BufFlt64+Tcount) = (dods_float64) *(bufflt+ipix);  
				    break;
		
				  case D_DOUBLE:
				    *(BufFlt64+Tcount) = (dods_float64) *(bufdbl+ipix);  
				    break;
		  
				  default:	
				    ErrMsgT("DSPArray: Invalid data conversion");
				    error = 1;
				    return false;
				}
		
				Tcount++;
			    }
			}
		    }
		    set_read_p(true);  
		    if((ConvertFrom == D_SHORT) || (ConvertFrom == D_LONG)){
			val2buf((void *)BufInt32);
			delete [] BufInt32;
		    }
		    else {
			val2buf((void *)BufFlt64);
			delete [] BufFlt64;
		    }

		}

	    }

    CloseDsp();

    error = 0;
    return false;
}
